/**
 * collectd - src/rrdd.c
 * Copyright (C) 2008 Florian octo Forster
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; only version 2 of the License is applicable.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 *
 * Authors:
 *   Florian octo Forster <octo at verplant.org>
 **/

#include "rrdd.h"

/*
 * Types
 */
struct listen_socket_s
{
  int fd;
  char path[PATH_MAX + 1];
};
typedef struct listen_socket_s listen_socket_t;

/*
 * Variables
 */
static listen_socket_t *listen_fds = NULL;
static size_t listen_fds_num = 0;

static int do_shutdown = 0;

static pthread_t queue_thread;

/* 
 * Functions
 */
static void *queue_thread_main (void *args) /* {{{ */
{
  while (do_shutdown == 0)
  {
    syslog (LOG_DEBUG, "queue_thread_main: Just woke up.");
    sleep (1);
  } /* while (do_shutdown == 0) */

  syslog (LOG_DEBUG, "queue_thread_main: Exiting.");

  return (NULL);
} /* }}} void *queue_thread_main */

static void sig_int_handler (int signal) /* {{{ */
{
  do_shutdown++;
} /* }}} void sig_int_handler */

static void sig_term_handler (int signal) /* {{{ */
{
  do_shutdown++;
} /* }}} void sig_term_handler */

static int open_listen_socket (const char *path) /* {{{ */
{
  int fd;
  struct sockaddr_un sa;
  listen_socket_t *temp;
  int status;

  temp = (listen_socket_t *) realloc (listen_fds,
      sizeof (listen_fds[0]) * (listen_fds_num + 1));
  if (temp == NULL)
  {
    syslog (LOG_ERR, "open_listen_socket: realloc failed.\n");
    return (-1);
  }
  listen_fds = temp;
  memset (listen_fds + listen_fds_num, 0, sizeof (listen_fds[0]));

  fd = socket (PF_UNIX, SOCK_STREAM, /* protocol = */ 0);
  if (fd < 0)
  {
    syslog (LOG_ERR, "open_listen_socket: socket(2) failed.\n");
    return (-1);
  }

  memset (&sa, 0, sizeof (sa));
  sa.sun_family = AF_UNIX;
  strncpy (sa.sun_path, path, sizeof (sa.sun_path) - 1);

  status = bind (fd, (struct sockaddr *) &sa, sizeof (sa));
  if (status != 0)
  {
    syslog (LOG_ERR, "open_listen_socket: bind(2) failed.\n");
    close (fd);
    return (-1);
  }
  
  listen_fds[listen_fds_num].fd = fd;
  strncpy (listen_fds[listen_fds_num].path, path,
      sizeof (listen_fds[listen_fds_num].path) - 1);
  listen_fds_num++;

  return (0);
} /* }}} int open_listen_socket */

static int close_listen_sockets (void) /* {{{ */
{
  size_t i;

  for (i = 0; i < listen_fds_num; i++)
  {
    close (listen_fds[i].fd);
    unlink (listen_fds[i].path);
  }

  free (listen_fds);
  listen_fds = NULL;
  listen_fds_num = 0;

  return (0);
} /* }}} int close_listen_sockets */

static void *listen_thread_main (void *args) /* {{{ */
{
  char buffer[4096];
  int status;

  status = open_listen_socket (RRDD_SOCK_PATH);
  if (status != 0)
  {
    syslog (LOG_ERR, "listen_thread_main: open_listen_socket failed.");
    return (NULL);
  }

  while (do_shutdown == 0)
  {
    syslog (LOG_DEBUG, "listen_thread_main: Just woke up.");

    status = recv (listen_fds[0].fd, buffer, sizeof (buffer),
        /* flags = */ 0);
    if (status == 0)
    {
      continue;
    }
    else if (status < 0)
    {
      if (errno == EINTR)
        continue;
      else
      {
        syslog (LOG_ERR, "listen_thread_main: recv failed.");
        continue;
      }
    }

    syslog (LOG_DEBUG, "listen_thread_main: Received %i bytes.\n", status);
  } /* while (do_shutdown == 0) */

  close_listen_sockets ();

  syslog (LOG_DEBUG, "listen_thread_main: Exiting.");

  return (NULL);
} /* }}} void *listen_thread_main */

static int daemonize (void) /* {{{ */
{
  pid_t child;
  int status;

  child = fork ();
  if (child < 0)
  {
    fprintf (stderr, "daemonize: fork(2) failed.\n");
    return (-1);
  }
  else if (child > 0)
  {
    return (1);
  }

  /* Change into the /tmp directory. */
  chdir ("/tmp");

  /* Become session leader */
  setsid ();

  /* Open the first three file descriptors to /dev/null */
  close (2);
  close (1);
  close (0);

  open ("/dev/null", O_RDWR);
  dup (0);
  dup (0);

  {
    struct sigaction sa;

    memset (&sa, 0, sizeof (sa));
    sa.sa_handler = sig_int_handler;
    sigaction (SIGINT, &sa, NULL);

    memset (&sa, 0, sizeof (sa));
    sa.sa_handler = sig_term_handler;
    sigaction (SIGINT, &sa, NULL);
  }

  openlog ("rrdd", LOG_PID, LOG_DAEMON);

  memset (&queue_thread, 0, sizeof (queue_thread));
  status = pthread_create (&queue_thread, /* attr = */ NULL,
      queue_thread_main, /* args = */ NULL);
  if (status != 0)
  {
    syslog (LOG_ERR, "daemonize: pthread_create failed.");
    return (-1);
  }

  return (0);
} /* }}} int daemonize */

static int cleanup (void) /* {{{ */
{
  do_shutdown++;

  pthread_join (queue_thread, /* return = */ NULL);

  closelog ();

  return (0);
} /* }}} int cleanup */

int main (int argc, char **argv)
{
  int status;

  printf ("%s by Florian Forster, Version %s\n",
      PACKAGE_NAME, PACKAGE_VERSION);

  status = daemonize ();
  if (status == 1)
  {
    struct sigaction sigchld;

    memset (&sigchld, 0, sizeof (sigchld));
    sigchld.sa_handler = SIG_IGN;
    sigaction (SIGCHLD, &sigchld, NULL);

    return (0);
  }
  else if (status != 0)
  {
    fprintf (stderr, "daemonize failed, exiting.\n");
    return (1);
  }

  listen_thread_main (NULL);

  cleanup ();

  return (0);
} /* int main */

/*
 * vim: set sw=2 sts=2 ts=8 et fdm=marker :
 */
